"""Module which provides functionality to perform fitting."""

import abc
from typing import List
import numpy as np

class AbstractFitting(abc.ABC):

    """Base class for fitting schemes."""

    supported_options = {}

    def __init__(self, **kwargs):
        self.set_options(**kwargs)

    @abc.abstractmethod
    def set_options(self, **kwargs):
        """Base method for setting options."""
        self.options = {}
        for key, value in kwargs.items():
            if key in self.supported_options:
                self.options[key] = value
            else:
                raise ValueError(f"Unsupported option: {key}")

        for option, default_value in self.supported_options.items():
            if option not in self.options:
                self.options[option] = default_value

    @abc.abstractmethod
    def fit(self, vectors: List[np.ndarray], target:np.ndarray):
        """Base method for computing new fitting coefficients."""
        return np.zeros(0)

    def linear_combination(self, vectors: List[np.ndarray],
            coefficients: np. ndarray) -> np.ndarray:
        """Given a set of vectors (or matrices) and the corresponding
        coefficients, build their linear combination."""
        result = np.zeros(vectors[0].shape, dtype=np.float64)
        for coeff, vector in zip(coefficients, vectors):
            result += vector*coeff
        return result

class LeastSquare(AbstractFitting):

    """Simple least square minimization fitting."""

    supported_options = {
        "regularization": 0.0,
    }

    def set_options(self, **kwargs):
        """Set options for least square minimization"""
        super().set_options(**kwargs)

        if self.options["regularization"] < 0 \
                or self.options["regularization"] > 100:
            raise ValueError("Unsupported value for regularization")

    def fit(self, vectors: List[np.ndarray], target: np.ndarray):
        """Given a set of vectors and a target return the fitting
        coefficients."""
        matrix = np.array(vectors).T
        a = matrix.T @ matrix
        b = matrix.T @ target
        if self.options["regularization"] > 0.0:
            a += np.identity(len(b))*self.options["regularization"]**2
        coefficients = np.linalg.solve(a, b)
        return np.array(coefficients, dtype=np.float64)

class QuasiTimeReversible(AbstractFitting):

    """Quasi time reversible fitting scheme."""

    supported_options = {
        "regularization": 0.0,
    }

    def set_options(self, **kwargs):
        """Set options for quasi time reversible fitting"""
        super().set_options(**kwargs)

        if self.options["regularization"] < 0 \
                or self.options["regularization"] > 100:
            raise ValueError("Unsupported value for regularization")

    def fit(self, vectors: List[np.ndarray], target: np.ndarray):
        """Given a set of vectors and a target return the fitting
        coefficients in a quasi time reversible scheme."""

        past_target = vectors[0]
        matrix = np.array(vectors[1:]).T

        q = matrix.shape[1]
        if q == 1:
            time_reversible_matrix = matrix
        elif q%2 == 0:
            time_reversible_matrix = matrix[:, :q//2] + matrix[:, :q//2-1:-1]
        else:
            time_reversible_matrix = matrix[:, :q//2+1] + matrix[:, :q//2-1:-1]

        a = time_reversible_matrix.T @ time_reversible_matrix
        b = time_reversible_matrix.T @ (target + past_target)

        if self.options["regularization"] > 0.0:
            a += np.identity(len(b))*self.options["regularization"]**2
        coefficients = np.linalg.solve(a, b)

        if q == 1:
            full_coefficients = np.concatenate(([-1.0], coefficients))
        elif q%2 == 0:
            full_coefficients = np.concatenate(([-1.0], coefficients,
                coefficients[::-1]))
        else:
            full_coefficients = np.concatenate(([-1.0], coefficients[:-1],
                2.0*coefficients[-1:], coefficients[-2::-1]))
        return np.array(full_coefficients, dtype=np.float64)
