import pytest
import os
import sys
import numpy as np

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import gext
import gext.descriptors
import gext.fitting
import gext.grassmann
import utils

SMALL = 1e-8
THRESHOLD = 5e-2

@pytest.mark.parametrize("datafile", ["urea.json", "glucose.json"])
@pytest.mark.parametrize("regularization", [0.0, 0.01, 0.05])
def test_least_square(datafile, regularization):

    # load test data from json file
    data = utils.load_json(f"tests/{datafile}")
    nelectrons = data["nelectrons"]
    natoms = data["trajectory"].shape[1]
    nbasis = data["overlaps"].shape[1]
    nframes = data["trajectory"].shape[0]

    # initialize an extrapolator
    extrapolator = gext.Extrapolator(nelectrons, nbasis, natoms,
        nsteps=nframes, fitting_regularization=regularization,
        fitting="leastsquare")

    # load data in the extrapolator
    for (coords, coeff, overlap) in zip(data["trajectory"],
            data["coefficients"], data["overlaps"]):
        extrapolator.load_data(coords, coeff, overlap)

    descriptors = extrapolator.descriptors.get(10)
    target = descriptors[-1]

    fitting_calculator = extrapolator.fitting_calculator

    # check if things are reasonable
    for start in range(0, 9):
        vectors = descriptors[start:-1]
        fit_coefficients = fitting_calculator.fit(vectors, target)
        fitted_target = fitting_calculator.linear_combination(vectors, fit_coefficients)
        error = np.linalg.norm(target - fitted_target, ord=np.inf)
        assert error < THRESHOLD

    # if we put the target in the vectors used for the fitting,
    # check that we get an error smaller than the regularization
    vectors = descriptors[:-1]
    vectors[0] = target
    fit_coefficients = fitting_calculator.fit(vectors, target)
    fitted_target = fitting_calculator.linear_combination(vectors, fit_coefficients)

    assert np.linalg.norm(target - fitted_target, ord=np.inf) < max(SMALL, regularization)

@pytest.mark.parametrize("datafile", ["urea.json", "glucose.json"])
@pytest.mark.parametrize("regularization", [0.0, 0.01, 0.05])
def test_quasi_time_reversible(datafile, regularization):

    # load test data from json file
    data = utils.load_json(f"tests/{datafile}")
    nelectrons = data["nelectrons"]
    natoms = data["trajectory"].shape[1]
    nbasis = data["overlaps"].shape[1]
    nframes = data["trajectory"].shape[0]

    # initialize an extrapolator
    extrapolator = gext.Extrapolator(nelectrons, nbasis, natoms,
        nsteps=nframes, fitting="qtr", fitting_regularization=regularization)

    # load data in the extrapolator
    for (coords, coeff, overlap) in zip(data["trajectory"],
            data["coefficients"], data["overlaps"]):
        extrapolator.load_data(coords, coeff, overlap)

    descriptors = extrapolator.descriptors.get(10)
    target = descriptors[-1]

    fitting_calculator = extrapolator.fitting_calculator

    # check if things are reasonable
    for start in range(0, 8):
        vectors = descriptors[start:-1]
        fit_coefficients = fitting_calculator.fit(vectors, target)
        fitted_target = fitting_calculator.linear_combination(vectors, fit_coefficients)
        error = np.linalg.norm(target - fitted_target, ord=np.inf)
        assert error < THRESHOLD

@pytest.mark.parametrize("datafile", ["urea.json", "glucose.json"])
def test_time_reversibility(datafile):

    # load test data from json file
    data = utils.load_json(f"tests/{datafile}")
    nelectrons = data["nelectrons"]
    natoms = data["trajectory"].shape[1]
    nbasis = data["overlaps"].shape[1]
    nframes = data["trajectory"].shape[0]

    # initialize an extrapolator
    extrapolator = gext.Extrapolator(nelectrons, nbasis, natoms,
        nsteps=nframes, fitting="qtr")

    # load data in the extrapolator
    for (coords, coeff, overlap) in zip(data["trajectory"],
            data["coefficients"], data["overlaps"]):
        extrapolator.load_data(coords, coeff, overlap)

    descriptors = extrapolator.descriptors.get(10)

    # we symmetrize the future and past targets (remember that it is
    # quasi time reversible, not exactly time reversible)
    target = descriptors[0] + descriptors[-1]
    descriptors[0] = target
    descriptors[-1] = target

    fitting_calculator = extrapolator.fitting_calculator

    # fit the future target
    fit_coefficients = fitting_calculator.fit(descriptors[:-1], descriptors[-1])
    fitted_target = fitting_calculator.linear_combination(descriptors[:-1], fit_coefficients)

    # fit the past target
    reversed_descriptors = list(reversed(descriptors))
    fit_coefficients = fitting_calculator.fit(reversed_descriptors[:-1], reversed_descriptors[-1])
    fitted_target_reverse = fitting_calculator.linear_combination(reversed_descriptors[:-1], fit_coefficients)

    # check the time reversibility
    assert np.linalg.norm(fitted_target - fitted_target_reverse, ord=np.inf) < SMALL

@pytest.mark.parametrize("datafile", ["urea.json", "glucose.json"])
@pytest.mark.parametrize("regularization", [0.001, 0.005, 0.01, 0.05, 0.1])
def test_square_qtr_formulation(datafile, regularization):

    # load test data from json file
    data = utils.load_json(f"tests/{datafile}")
    nelectrons = data["nelectrons"]
    natoms = data["trajectory"].shape[1]
    nbasis = data["overlaps"].shape[1]
    nframes = data["trajectory"].shape[0]

    # initialize an extrapolator
    extrapolator = gext.Extrapolator(nelectrons, nbasis, natoms,
        nsteps=nframes, fitting="qtr", fitting_regularization=regularization)

    # load data in the extrapolator
    for (coords, coeff, overlap) in zip(data["trajectory"],
            data["coefficients"], data["overlaps"]):
        extrapolator.load_data(coords, coeff, overlap)

    descriptors = extrapolator.descriptors.get(10)
    target = descriptors[-1]

    fitting_calculator = extrapolator.fitting_calculator
    alt_fitting_calculator = utils.AlternativeQuasiTimeReversible()
    alt_fitting_calculator.set_options(regularization=regularization)

    # check if the two fittings give the same coefficients
    for start in range(0, 8):
        vectors = descriptors[start:-1]
        fit_coefficients = fitting_calculator.fit(vectors, target)
        alt_fit_coefficients = alt_fitting_calculator.fit(vectors, target)
        assert np.linalg.norm(fit_coefficients - alt_fit_coefficients, ord=np.inf) < 1e-4

@pytest.mark.parametrize("datafile", ["urea.json", "glucose.json"])
@pytest.mark.parametrize("regularization", [0.001, 0.005, 0.01, 0.05, 0.1])
def test_square_ls_formulation(datafile, regularization):

    # load test data from json file
    data = utils.load_json(f"tests/{datafile}")
    nelectrons = data["nelectrons"]
    natoms = data["trajectory"].shape[1]
    nbasis = data["overlaps"].shape[1]
    nframes = data["trajectory"].shape[0]

    # initialize an extrapolator
    extrapolator = gext.Extrapolator(nelectrons, nbasis, natoms,
        nsteps=nframes, fitting="leastsquare", fitting_regularization=regularization)

    # load data in the extrapolator
    for (coords, coeff, overlap) in zip(data["trajectory"],
            data["coefficients"], data["overlaps"]):
        extrapolator.load_data(coords, coeff, overlap)

    descriptors = extrapolator.descriptors.get(10)
    target = descriptors[-1]

    fitting_calculator = extrapolator.fitting_calculator
    alt_fitting_calculator = utils.AlternativeLeastSquare()
    alt_fitting_calculator.set_options(regularization=regularization)

    # check if the two fittings give the same coefficients
    for start in range(0, 8):
        vectors = descriptors[start:-1]
        fit_coefficients = fitting_calculator.fit(vectors, target)
        alt_fit_coefficients = alt_fitting_calculator.fit(vectors, target)
        assert np.linalg.norm(fit_coefficients - alt_fit_coefficients, ord=np.inf) < 1e-4
